using System;
using System.Collections.Generic;

namespace AmortizationCalculator
{
    /// <summary>A simplified version of the standard List&lt;T&gt; with added events.</summary>
    /// <typeparam name="T"></typeparam>
    public class AbbrevList<T> : IEnumerable<T>
    {
        private List<T> _list = new List<T>();

        public T this[int index]
        {
            get { return _list[index]; }
            set
            {
                _list[index] = value;
                OnModified(index);
            }
        }

        public void Add(T item)
        {
            _list.Add(item);
            int index = _list.IndexOf(item);
            OnAdded(index);
        }

        public int Count
        {
            get { return _list.Count; }
        }

        public void RemoveAt(int index)
        {
            _list.RemoveAt(index);
            OnRemoved(index);
        }

        #region IEnumerable<T> Members
        public IEnumerator<T> GetEnumerator()
        {
            return _list.GetEnumerator();
        }

        System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
        {
            return GetEnumerator();
        }
        #endregion

        #region Events
        /// <summary>in practice, all items added are null</summary>
        public event IntEventHandler Added;
        protected virtual void OnAdded(int number)
        {
            if (Added != null)
                Added(this, new IntEventArgs(number));
        }

        public event IntEventHandler Modified;
        protected virtual void OnModified(int number)
        {
            if (Modified != null)
                Modified(this, new IntEventArgs(number));
        }

        public event IntEventHandler Removed;
        protected virtual void OnRemoved(int number)
        {
            if (Removed != null)
                Removed(this, new IntEventArgs(number));
        }
        #endregion
    }
}
