using System;
using System.Drawing;
using System.IO;
using System.Windows.Forms;
using System.Xml.Serialization;
using AmortizationCalculator;

namespace WinForm
{
    public partial class AmortizationForm : Form
    {
        private AmortizationModel _model = new AmortizationModel();
        public static AmortCalcFormSettings Settings = new AmortCalcFormSettings();

        private Panel headerPanel = new Panel();
        private Panel rowPanel = new Panel();

        private int formHeightWithoutRows;
        private string origText;

        PaymentsPeriodComboBoxItem defaultSelectedPaymentPeriod;

        #region Public c'tor
        public AmortizationForm()
        {
            InitializeComponent();

            origText = this.Text;

            #region set up expanding/contracting panels
            createPanels();
            createHeader();
            formHeightWithoutRows = this.Height;
            #endregion

            #region subscribe to model events
            _model.PaymentChanged += new EventHandler(_model_PaymentChanged);
            _model.LoanTermsChanged += new EventHandler(_model_LoanTermsChanged);
            _model.InvalidLoanData += new EventHandler(_model_InvalidLoanData);

            _model.Added += new IntArrayEventHandler(rowsAdded);
            _model.Modified += new IntArrayEventHandler(rowsModified);
            _model.Removed += new IntArrayEventHandler(rowsRemoved);
            #endregion

            #region set up payment freq combobox
            defaultSelectedPaymentPeriod = new PaymentsPeriodComboBoxItem(
                PaymentsPeriod.Monthly, "Monthly");
            paymentFrequencyCb.Items.Add(new PaymentsPeriodComboBoxItem(
                PaymentsPeriod.Annually, "Annually"));
            paymentFrequencyCb.Items.Add(new PaymentsPeriodComboBoxItem(
                PaymentsPeriod.SemiAnnually, "Semi-Annually (2x/yr.)"));
            paymentFrequencyCb.Items.Add(new PaymentsPeriodComboBoxItem(
                PaymentsPeriod.EveryFourMonths, "Every 4 Months"));
            paymentFrequencyCb.Items.Add(new PaymentsPeriodComboBoxItem(
                PaymentsPeriod.Quarterly, "Quarterly"));
            paymentFrequencyCb.Items.Add(new PaymentsPeriodComboBoxItem(
                PaymentsPeriod.BiMonthly, "Bi-Monthly (every 2 mo.)"));
            paymentFrequencyCb.Items.Add(defaultSelectedPaymentPeriod);

            // init
            _userPaymentQtyInput = numberPaymentsTb;
            #endregion

            #region create/import settings
            TextReader reader = null;
            try
            {
                XmlSerializer serializer = new XmlSerializer(typeof(AmortCalcFormSettings));
                reader = new StreamReader(AmortCalcFormSettings.SETTINGS_FILE);
                Settings = (AmortCalcFormSettings)serializer.Deserialize(reader);
            }
            catch { }
            finally
            {
                if (reader != null)
                    reader.Close();
            }
            Settings.AllowSave = true;

            if (Settings.DoSave)
                open(AmortCalcFormSettings.STATE_FILE);
            else
                resetForm();
            Settings.SaveState();
            #endregion

            // if none set by this point, set default
            if (paymentFrequencyCb.SelectedIndex == -1)
                paymentFrequencyCb.SelectedItem = defaultSelectedPaymentPeriod;
        }

        private void createPanels()
        {
            int headerX = 12;
            int headerY = 183;
            int headerHeight = 23;

            headerPanel.Location = new Point(headerX, headerY);
            headerPanel.AutoScroll = true;
            headerPanel.Size = new Size(638, headerHeight);
            headerPanel.Visible = false;
            toolStripContainer.ContentPanel.Controls.Add(headerPanel);

            rowPanel.Anchor = AnchorStyles.Bottom | AnchorStyles.Top | AnchorStyles.Left | AnchorStyles.Right;
            rowPanel.Location = new Point(headerX, headerY + headerHeight);
            rowPanel.AutoScroll = true;
            rowPanel.Size = new Size(673, 0);
            rowPanel.Click += new EventHandler(panel_Focus);
            toolStripContainer.ContentPanel.Controls.Add(rowPanel);
        }

        private void createHeader()
        {
            Label paymentNumberColLbl = new Label();
            Label paymentDateColLbl = new Label();
            Label interestColLbl = new Label();
            Label principalColLbl = new Label();
            Label paymentColLbl = new Label();
            Label extraPaymentColLbl = new Label();
            Label cumInterestColLbl = new Label();
            Label cumPrincipalColLbl = new Label();
            Label princEndColLbl = new Label();

            headerPanel.Controls.Add(paymentNumberColLbl);
            headerPanel.Controls.Add(paymentDateColLbl);
            headerPanel.Controls.Add(interestColLbl);
            headerPanel.Controls.Add(principalColLbl);
            headerPanel.Controls.Add(paymentColLbl);
            headerPanel.Controls.Add(extraPaymentColLbl);
            headerPanel.Controls.Add(cumInterestColLbl);
            headerPanel.Controls.Add(cumPrincipalColLbl);
            headerPanel.Controls.Add(princEndColLbl);

            Font font = new Font("Microsoft Sans Serif", 8.25F, FontStyle.Bold, GraphicsUnit.Point, ((byte)(0)));

            int x, y;

            // new row
            x = Row.RowX0;
            y = Row.RowY;
            // paymentNumberColLbl
            paymentNumberColLbl.AutoSize = false;
            paymentNumberColLbl.TextAlign = ContentAlignment.BottomCenter;
            paymentNumberColLbl.Size = new Size(Row.RowX1 - Row.ColBuffer, Row.LabelHeight);
            paymentNumberColLbl.Font = font;
            paymentNumberColLbl.Location = new Point(x, y);
            paymentNumberColLbl.Text = "#";
            paymentNumberColLbl.Click += new EventHandler(paymentNumberColLbl_Click);
            // paymentDateColLbl
            x += Row.RowX1;
            paymentDateColLbl.AutoSize = false;
            paymentDateColLbl.TextAlign = ContentAlignment.BottomCenter;
            paymentDateColLbl.Size = new Size(Row.RowX2 - Row.ColBuffer, Row.LabelHeight);
            paymentDateColLbl.Font = font;
            paymentDateColLbl.Location = new Point(x, y);
            paymentDateColLbl.Text = "Date";
            paymentDateColLbl.Click += new EventHandler(paymentDateColLbl_Click);
            // interestColLbl
            x += Row.RowX2;
            interestColLbl.AutoSize = false;
            interestColLbl.TextAlign = ContentAlignment.BottomCenter;
            interestColLbl.Size = new Size(Row.RowX3 - Row.ColBuffer, Row.LabelHeight);
            interestColLbl.Font = font;
            interestColLbl.Location = new Point(x, y);
            interestColLbl.Text = "Interest";
            interestColLbl.Click += new EventHandler(this.interestColLbl_Click);
            // principalColLbl
            x += Row.RowX3;
            principalColLbl.AutoSize = false;
            principalColLbl.TextAlign = ContentAlignment.BottomCenter;
            principalColLbl.Size = new Size(Row.RowX4 - Row.ColBuffer, Row.LabelHeight);
            principalColLbl.Font = font;
            principalColLbl.Location = new Point(x, y);
            principalColLbl.Text = "Principal";
            principalColLbl.Click += new EventHandler(principalColLbl_Click);
            // paymentColLbl
            x += Row.RowX4;
            paymentColLbl.AutoSize = false;
            paymentColLbl.TextAlign = ContentAlignment.BottomCenter;
            paymentColLbl.Size = new Size(Row.RowX5 - Row.ColBuffer, Row.LabelHeight);
            paymentColLbl.Font = font;
            paymentColLbl.Location = new Point(x, y);
            paymentColLbl.Text = "Payment";
            paymentColLbl.Click += new EventHandler(paymentColLbl_Click);
            // extraPaymentColLbl
            x += Row.RowX5;
            extraPaymentColLbl.AutoSize = true;
            extraPaymentColLbl.Font = font;
            extraPaymentColLbl.Location = new Point(x, y);
            extraPaymentColLbl.Text = "Extra Payment";
            extraPaymentColLbl.Click += new EventHandler(extraPaymentColLbl_Click);
            // cumInterestColLbl
            x += Row.RowX6;
            cumInterestColLbl.AutoSize = false;
            cumInterestColLbl.TextAlign = ContentAlignment.BottomCenter;
            cumInterestColLbl.Size = new Size(Row.RowX7 - Row.ColBuffer, Row.LabelHeight);
            cumInterestColLbl.Font = font;
            cumInterestColLbl.Location = new Point(x, y);
            cumInterestColLbl.Text = "Cum Int";
            cumInterestColLbl.Click += new EventHandler(cumInterestColLbl_Click);
            // cumPrincipalColLbl
            x += Row.RowX7;
            cumPrincipalColLbl.AutoSize = false;
            cumPrincipalColLbl.TextAlign = ContentAlignment.BottomCenter;
            cumPrincipalColLbl.Size = new Size(Row.RowX8 - Row.ColBuffer, Row.LabelHeight);
            cumPrincipalColLbl.Font = font;
            cumPrincipalColLbl.Location = new Point(x, y);
            cumPrincipalColLbl.Text = "Cum Princ";
            cumPrincipalColLbl.Click += new EventHandler(cumPrincipalColLbl_Click);
            // princEndColLbl
            x += Row.RowX8;
            princEndColLbl.AutoSize = true;
            princEndColLbl.Font = font;
            princEndColLbl.Location = new Point(x, y);
            princEndColLbl.Text = "Princ Balance";
            princEndColLbl.Click += new EventHandler(princEndColLbl_Click);
        }
        #endregion

        #region methods for model events
        private void _model_PaymentChanged(object sender, EventArgs e)
        {
            paymentCalcLbl.Text = String.Format("{0:c}", _model.PaymentAmount);
        }
        private void _model_LoanTermsChanged(object sender, EventArgs e)
        {
            principalTb.Text
                = (_model.Principal > 0)
                ? String.Format("{0:c}", _model.Principal)
                : "";
            aprTb.Text
                = (_model.Apr > 0)
                ? _model.Apr.ToString()
                : "";
            startDateDtp.Value = _model.StartDate;
            for (int i = 0; i < paymentFrequencyCb.Items.Count; i++)
            {
                PaymentsPeriodComboBoxItem item
                    = (PaymentsPeriodComboBoxItem)paymentFrequencyCb.Items[i];
                if (item.Period == _model.PaymentsPerYear)
                {
                    paymentFrequencyCb.SelectedIndex = i;
                    break;
                }
            }
            numberPaymentsTb.Text
                = (_model.QtyPayments > 0)
                ? _model.QtyPayments.ToString()
                : "";
            // make sure qty years is populated
            numberPaymentsTb_Submitted(sender, e);

            totalInterestCalcLbl.Text = String.Format("{0:c}", _model.TotalInterestPaid);
            totalPrincipalCalcLbl.Text = String.Format("{0:c}", _model.TotalPrincipalPaid);
            totalRepaidCalcLbl.Text = String.Format("{0:c}", _model.TotalRepaid);
            finalNumberPaymentsCalcLbl.Text = _model.QtyPaymentsWithExtraPayments.ToString();
        }
        private void _model_InvalidLoanData(object sender, EventArgs e)
        {
            rowPanel.Enabled = false;
            resetPaymentStats();
        }

        private bool isValid(double currency)
        {
            return !(
                double.IsNaN(currency) ||
                double.IsInfinity(currency) ||
                double.IsNegativeInfinity(currency) ||
                currency == 0);
        }
        private void resetPaymentStats()
        {
            paymentCalcLbl.Text = "--";
            totalInterestCalcLbl.Text = "--";
            totalPrincipalCalcLbl.Text = "--";
            totalRepaidCalcLbl.Text = "--";
            finalNumberPaymentsCalcLbl.Text = "--";
        }

        private void rowsAdded(object sender, IntArrayEventArgs e)
        {
            rowPanel.Enabled = true;
            expandBorders();

            foreach (int id in e.Numbers)
            {
                Row r = new Row(id);
                SinglePayment payment = _model[id];

                if (payment != null)
                    r.SetAll(payment);

                // set all fields
                r.Click += new EventHandler(panel_Focus);
                r.TextBoxChanged += new IntEventHandler(rowTextBoxChanged);
                r.Location = new Point(0, id * Row.RowHeight + rowPanel.AutoScrollPosition.Y);

                r.Visible = false;
                rowPanel.Controls.Add(r);
                r.Visible = true;
            }

            headerPanel.Visible = true;
            rowPanel.Visible = true;
        }

        private void rowTextBoxChanged(object sender, IntEventArgs e)
        {
            int id = e.Number;
            if (id >= _model.Count)
                return;

            SinglePayment payment = _model[id];
            Row r = (Row)rowPanel.Controls[id];
            double extraPayment = parseCurrency(r.ExtraPaymentText);

            if (payment.ExtraPayment == extraPayment)
                return;

            payment.SetExtraPayment(extraPayment);
            saveState();
        }

        private void rowsModified(object sender, IntArrayEventArgs e)
        {
            rowPanel.Enabled = true;
            foreach (int id in e.Numbers)
            {
                SinglePayment payment = _model[id];

                if (payment == null)
                    continue;

                Row r = (Row)rowPanel.Controls[id];
                r.SetAll(payment);
            }
        }

        private void rowsRemoved(object sender, IntArrayEventArgs e)
        {
            if (rowPanel.Controls.Count == 0)
                return;

            foreach (int id in e.Numbers)
                rowPanel.Controls[id].Visible = false;
            
            expandBorders();

            foreach (int id in e.Numbers)
                rowPanel.Controls.RemoveAt(id);

            if (rowPanel.Controls.Count == 0)
            {
                headerPanel.Visible = false;
                rowPanel.Visible = false;
            }
        }

        /// <summary>call this method after the model has been modified and before the rowPanel has been</summary>
        private void expandBorders()
        {
            rowPanel.Enabled = true;

            if (WindowState == FormWindowState.Maximized ||
                WindowState == FormWindowState.Minimized ||
                _model.Count == rowPanel.Controls.Count)
                return;

            int sign, pixels;

            int rowsHeightAvail = this.Height - formHeightWithoutRows;
            int rowsHeightUsed = rowPanel.Controls.Count * Row.RowHeight;
            int rowsHeightNeeded = _model.Count * Row.RowHeight;

            // if model count is less, then rowPanels is about to remove an item
            if (_model.Count < rowPanel.Controls.Count && rowsHeightAvail > rowsHeightNeeded)
            {
                sign = -1;

                // use available space
                pixels = Math.Min(rowsHeightNeeded - rowsHeightAvail, 0);
            }
            // if model count is more, then rowPanels is about to add an item.
            else
            {
                sign = 1;

                // use available space
                int avail = Math.Max(rowsHeightNeeded - rowsHeightAvail, 0);

                // don't go off the screen
                int maxFormHeight = Settings.MaxRowHeight * Row.RowHeight + formHeightWithoutRows;
                int currMaxFormHeight = Math.Min(Screen.PrimaryScreen.Bounds.Height - this.Location.Y, maxFormHeight);
                // don't exceed max allowed
                pixels = Math.Min(Math.Max(currMaxFormHeight, this.Height), avail + this.Height) - this.Height;
            }

            float curr = 0;
            float buffer = sign * 0.4F;

            int origHeight = this.Height;

            while (Math.Abs(buffer + curr) < sign * pixels)
            {
                curr -= (curr - pixels) / 5.0F;
                this.Size = new Size(this.Width, (int)(origHeight + curr));
            }
            this.Size = new Size(this.Width, (int)(origHeight + pixels));
        }
        #endregion

        #region loan terms changed
        private void principalTb_Submitted(object sender, EventArgs e)
        {
            double newPrincipal = parseCurrency(principalTb.Text);
            if (newPrincipal == _model.Principal)
                return;
            _model.Principal = newPrincipal;
            saveState();
        }

        private void aprTb_Submitted(object sender, EventArgs e)
        {
            string apr = aprTb.Text.Trim(new char[] { '%' });
            if (apr == "")
            {
                _model.Apr = 0;
                return;
            }

            double dApr;
            double newApr = (Double.TryParse(apr, out dApr)) ? dApr : 0;
            if (newApr == _model.Apr)
                return;
            _model.Apr = newApr;

            saveState();
        }

        private void startDateDtp_ValueChanged(object sender, EventArgs e)
        {
            DateTime newDate = startDateDtp.Value;
            if (newDate == _model.StartDate)
                return;
            _model.StartDate = newDate;
            saveState();
        }

        private void paymentFrequencyCb_SelectedIndexChanged(object sender, EventArgs e)
        {
            PaymentsPeriod newPaymentFrequency = getPaymentFrequency();
            if (newPaymentFrequency == _model.PaymentsPerYear)
                return;
            _model.PaymentsPerYear = newPaymentFrequency;

            saveState();

            // re-calc num payments or num yrs
            if (_userPaymentQtyInput == numberPaymentsTb)
                numberPaymentsTb_Submitted(null, null);
            else
                numberYearsTb_Submitted(null, null);
        }

        /// <summary>num of payments or num of years entered by user</summary>
        private TextBox _userPaymentQtyInput;

        private void numberPaymentsTb_Submitted(object sender, EventArgs e)
        {
            string numberPayments = numberPaymentsTb.Text.Trim();
            if (numberPayments == "")
            {
                _model.QtyPayments = 0;
                return;
            }

            int iNumberPayments = 0;
            bool success = int.TryParse(numberPayments, out iNumberPayments);
            if (!success || iNumberPayments <= 0)
            {
                _model.QtyPayments = 0;
                return;
            }
            if (iNumberPayments > 960)
            {
                numberPaymentsTb.Text = "ERROR. Too large";
                _model.QtyPayments = 0;
                return;
            }

            bool doSaveState;
            if (iNumberPayments == _model.QtyPayments)
            {
                doSaveState = false;
            }
            else
            {
                doSaveState = true;
                _model.QtyPayments = iNumberPayments;
                _userPaymentQtyInput = numberPaymentsTb;
            }

            numberYearsTb.Text = qtyYearsFormatted();

            if (doSaveState)
                saveState();
        }

        private string qtyYearsFormatted()
        {
            // don't combine, or it could trim zeros to the left of the decimal as well
            string retStr = String.Format("{0:0.00}", _model.QtyYears).TrimEnd('0');
            retStr = retStr.TrimEnd('.');
            return retStr;
        }

        private void numberYearsTb_Submitted(object sender, EventArgs e)
        {
            string numberYears = numberYearsTb.Text.Trim();
            if (numberYears == "")
            {
                _model.QtyPayments = 0;
                return;
            }

            if (numberYears == qtyYearsFormatted())
                return;

            double dNumberYears = 0;
            bool success = double.TryParse(numberYears, out dNumberYears);
            if (!success || dNumberYears <= 0)
            {
                _model.QtyPayments = 0;
                return;
            }
            if (dNumberYears > 80)
            {
                numberYearsTb.Text = "ERROR. Too large";
                _model.QtyPayments = 0;
                return;
            }

            double dQtyPayments = dNumberYears * (int)getPaymentFrequency();
            int qtyPayments = (int)Math.Ceiling(dQtyPayments);
            int oldQtyPayments = _model.QtyPayments;
            _model.QtyPayments = qtyPayments;
            if (oldQtyPayments != _model.QtyPayments)
                _userPaymentQtyInput = numberYearsTb;

            numberPaymentsTb.Text = _model.QtyPayments.ToString();

            saveState();
        }

        private double parseCurrency(string currency)
        {
            if (currency == null)
                return 0;
            currency = currency.Trim(new char[] { '$' });
            if (currency == "")
                return 0;

            double dCurrency;
            bool success = Double.TryParse(currency, out dCurrency);
            if (!success)
                return 0;
            return dCurrency;
        }

        private PaymentsPeriod getPaymentFrequency()
        {
            return ((PaymentsPeriodComboBoxItem)paymentFrequencyCb.SelectedItem).Period;
        }
        #endregion

        #region label click info
        #region loan terms click info
        private void principalLbl_Click(object sender, EventArgs e)
        {
            explanationTsLbl.Text = "The amount borrowed.";
        }
        private void aprLbl_Click(object sender, EventArgs e)
        {
            explanationTsLbl.Text = "Annual Percentage Rate. The rate at which interest is calculated annually.";
        }

        private void startDataLbl_Click(object sender, EventArgs e)
        {
            explanationTsLbl.Text = "The date of the first payment.";
        }

        private void paymentFrequencyLbl_Click(object sender, EventArgs e)
        {
            explanationTsLbl.Text = "How many times per year a payment is made.";
        }

        private void numberPaymentsLbl_Click(object sender, EventArgs e)
        {
            explanationTsLbl.Text = "Total number of payments made to pay off the loan. This will automatically fill in \"Number of years\"";
        }

        private void numberYearsLbl_Click(object sender, EventArgs e)
        {
            explanationTsLbl.Text = "Number of years of payments to pay off the loan. This will automatically fill in \"Number of payments\"";
        }
        #endregion
        #region payment click info
        private void paymentLbl_Click(object sender, EventArgs e)
        {
            explanationTsLbl.Text = "The payment amount.";
        }

        private void totalInterestLbl_Click(object sender, EventArgs e)
        {
            explanationTsLbl.Text = "The total amount of interest paid.";
        }

        private void totalPrincipalLbl_Click(object sender, EventArgs e)
        {
            explanationTsLbl.Text = "The total amount of principal paid.";
        }

        private void totalRepaidLbl_Click(object sender, EventArgs e)
        {
            explanationTsLbl.Text = "The total amount of paid. (interest + principal)";
        }

        private void finalNumberPaymentsLbl_Click(object sender, EventArgs e)
        {
            explanationTsLbl.Text = "The number of payments made. Including extra payments.";
        }
        #endregion
        #region col headers
        private void paymentNumberColLbl_Click(object sender, EventArgs e)
        {
            explanationTsLbl.Text = "Which payment this is. E.g.: 1st, 2nd";
        }

        private void paymentDateColLbl_Click(object sender, EventArgs e)
        {
            explanationTsLbl.Text = "Payment due date.";
        }

        private void interestColLbl_Click(object sender, EventArgs e)
        {
            explanationTsLbl.Text = "How much of this payment is interest. (The rest will be principal.)";
        }

        private void principalColLbl_Click(object sender, EventArgs e)
        {
            explanationTsLbl.Text = "How much of this payment is principal. (The rest will be interest.)";
        }

        private void paymentColLbl_Click(object sender, EventArgs e)
        {
            explanationTsLbl.Text = "The payment amount.";
        }

        private void extraPaymentColLbl_Click(object sender, EventArgs e)
        {
            explanationTsLbl.Text = "Optional extra payments toward paying of the loan early. This is applied to the principal.";
        }

        private void cumInterestColLbl_Click(object sender, EventArgs e)
        {
            explanationTsLbl.Text = "Cumulative interest to date.";
        }

        private void cumPrincipalColLbl_Click(object sender, EventArgs e)
        {
            explanationTsLbl.Text = "Cumulative principal to date.";
        }

        private void princEndColLbl_Click(object sender, EventArgs e)
        {
            explanationTsLbl.Text = "The remaining balance of the original principal.";
        }
        #endregion
        #endregion

        private void panel_Focus(object sender, EventArgs e)
        {
            rowPanel.Focus();
        }

        #region menus and tools
        #region menus
        #region file
        private void newToolStripMenuItem1_Click(object sender, EventArgs e)
        {
            resetForm();
        }

        private void openToolStripMenuItem1_Click(object sender, EventArgs e)
        {
            openFileDialog.FileName = "";
            if (openFileDialog.ShowDialog() == DialogResult.OK)
                fileName = openFileDialog.FileName;
            open(fileName);
        }

        private void saveToolStripMenuItem1_Click(object sender, EventArgs e)
        {
            if (File.Exists(fileName))
                save();
            else
                saveAsToolStripMenuItem1_Click(sender, e);
        }

        private void saveAsToolStripMenuItem1_Click(object sender, EventArgs e)
        {
            saveFileDialog.FileName = "";
            if (saveFileDialog.ShowDialog() == DialogResult.OK)
            {
                fileName = saveFileDialog.FileName;
                save();
            }
        }

        private void exportToExcelToolStripMenuItem_Click(object sender, EventArgs e)
        {
            excelSaveFileDialog.FileName
                = (fileName != null)
                ? excelSaveFileDialog.FileName
                = Path.GetFileNameWithoutExtension(fileName) : "";
            if (excelSaveFileDialog.ShowDialog() == DialogResult.OK)
                export(excelSaveFileDialog.FileName);
        }

        private void exitToolStripMenuItem1_Click(object sender, EventArgs e)
        {
            Close();
        }
        #endregion

        #region tools
        private Options _options = new Options();
        private void optionsToolStripMenuItem1_Click(object sender, EventArgs e)
        {
            _options.ShowDialog();
        }
        #endregion

        #region help
        private HowToUse _howToUse = new HowToUse();
        private void howToUseToolStripMenuItem1_Click(object sender, EventArgs e)
        {
            _howToUse.ShowDialog();
        }

        private About _about = new About();
        private void aboutToolStripMenuItem1_Click(object sender, EventArgs e)
        {
            _about.ShowDialog();
        }
        #endregion
        #endregion

        #region tool strip
        private void newToolStripButton_Click(object sender, EventArgs e)
        {
            resetForm();
        }

        private void openToolStripButton_Click(object sender, EventArgs e)
        {
            openToolStripMenuItem1_Click(sender, e);
        }

        private void saveToolStripButton_Click(object sender, EventArgs e)
        {
            saveToolStripMenuItem1_Click(sender, e);
        }

        private void exportToExcelToolStripButton_Click(object sender, EventArgs e)
        {
            exportToExcelToolStripMenuItem_Click(sender, e);
        }

        private void helpToolStripButton_Click(object sender, EventArgs e)
        {
            howToUseToolStripMenuItem1_Click(sender, e);
        }
        #endregion
        #endregion

        #region open, save, reset, export, et al
        private void resetForm()
        {
            savePrompt();
            fileName = null;
            this.Text = origText;

            _model.Principal = 0;
            _model.Apr = 0;
            _model.StartDate = DateTime.Now;
            _model.PaymentsPerYear = PaymentsPeriod.Monthly;
            _model.QtyPayments = 0;
            numberYearsTb.Text = "";

            foreach (SinglePayment payment in _model)
                payment.ExtraPayment = 0;
            saveState();
        }

        private void open(string file)
        {
            savePrompt();

            TextReader reader = null;
            AmortizationBluePrint bluePrint;
            try
            {
                XmlSerializer serializer = new XmlSerializer(typeof(AmortizationBluePrint));
                reader = new StreamReader(file);
                bluePrint = (AmortizationBluePrint)serializer.Deserialize(reader);
            }
            catch { return; }
            finally
            {
                if (reader != null)
                    reader.Close();
            }

            _model.SetBluePrint(bluePrint);

            // different than using saveState because nothing
            // has changed so don't want asterisk
            if (file != AmortCalcFormSettings.STATE_FILE)
            {
                save(AmortCalcFormSettings.STATE_FILE);
                fileName = file;
            }

            setFormText(file);
        }

        private string fileName = null;
        private void save()
        {
            save(fileName);
        }
        private void saveState()
        {
            if (fileName != null && !this.Text.EndsWith("*"))
                this.Text += " *";
            save(AmortCalcFormSettings.STATE_FILE);
        }
        private void save(string file)
        {
            if (!File.Exists(file))
                File.Create(file).Close();

            XmlSerializer serializer = new XmlSerializer(typeof(AmortizationBluePrint));
            TextWriter writer = new StreamWriter(file);
            serializer.Serialize(writer, _model.GetBluePrint());
            writer.Close();

            setFormText(file);
        }

        private void setFormText(string file)
        {
            if (file == null || file == AmortCalcFormSettings.STATE_FILE)
                return;
            this.Text = origText + " - " + Path.GetFileName(file);
        }

        private void export(string fileName)
        {
            ExcelSpreadsheetGenerator generator = new ExcelSpreadsheetGenerator(_model);
            generator.CreateFile(fileName);
        }

        private void AmortizationForm_FormClosing(object sender, FormClosingEventArgs e)
        {
            savePrompt();
        }

        private void savePrompt()
        {
            if (fileName == null || !this.Text.EndsWith("*"))
                return;

            if (MessageBox.Show("Save '" + Path.GetFileName(fileName) + "'?", "Save?", MessageBoxButtons.YesNo) == DialogResult.Yes)
                save();
        }
        #endregion

        #region drag + drop
        private void AmortizationForm_DragEnter(object sender, DragEventArgs e)
        {
            e.Effect
                = (e.Data.GetDataPresent(DataFormats.FileDrop))
                ? DragDropEffects.All
                : DragDropEffects.None;
        }

        private void AmortizationForm_DragDrop(object sender, DragEventArgs e)
        {
            foreach (string file in (string[])e.Data.GetData(DataFormats.FileDrop))
            {
                if (Path.GetExtension(file).ToLower() == ".azc")
                {
                    open(file);
                    break;
                }
            }
        }
        #endregion
    }
    public class PaymentsPeriodComboBoxItem
    {
        public PaymentsPeriod Period;
        public string Text;
        public PaymentsPeriodComboBoxItem() { }
        public PaymentsPeriodComboBoxItem(PaymentsPeriod period, string text)
        {
            Period = period;
            Text = text;
        }
        public override string ToString()
        {
            return Text;
        }
    }
    public class Row : Panel
    {
        #region constants
        public const int RowWidth = 625;
        public const int RowHeight = 26;

        public const int ColBuffer = 6;

        public const int RowY = 3;

        public const int RowX0 = 0;
        public const int RowX1 = 31;
        public const int RowX2 = 59;
        public const int RowX3 = 67;
        public const int RowX4 = 67;
        public const int RowX5 = 67;
        public const int RowX6 = 86;
        public const int RowX7 = 88;
        public const int RowX8 = 82;

        public const int LabelHeight = 13;
        #endregion

        int _id;

        private Label paymentNumberLbl = new Label();
        private Label paymentDateLbl = new Label();
        private Label interestLbl = new Label();
        private Label principalLbl = new Label();
        private Label paymentLbl = new Label();
        private EntryTextBox extraPaymentTb = new EntryTextBox();
        private Label cumInterestLbl = new Label();
        private Label cumPrincipalLbl = new Label();
        private Label princEndLbl = new Label();

        #region Public Properties
        //public string PaymentNumberText
        //{
        //    get { return paymentNumberLbl.Text; }
        //    set { paymentNumberLbl.Text = value; }
        //}
        //public string PaymentDateText
        //{
        //    get { return paymentDateLbl.Text; }
        //    set { paymentDateLbl.Text = value; }
        //}
        //public string InterestText
        //{
        //    get { return interestLbl.Text; }
        //    set { interestLbl.Text = value; }
        //}
        //public string PrincipalText
        //{
        //    get { return principalLbl.Text; }
        //    set { principalLbl.Text = value; }
        //}
        //public string PaymentText
        //{
        //    get { return paymentLbl.Text; }
        //    set { paymentLbl.Text = value; }
        //}
        public string ExtraPaymentText
        {
            get { return extraPaymentTb.Text; }
            set { extraPaymentTb.Text = value; }
        }
        //public string CumInterestText
        //{
        //    get { return cumInterestLbl.Text; }
        //    set { cumInterestLbl.Text = value; }
        //}
        //public string CumPrincipalText
        //{
        //    get { return cumPrincipalLbl.Text; }
        //    set { cumPrincipalLbl.Text = value; }
        //}
        //public string PrincEndText
        //{
        //    get { return princEndLbl.Text; }
        //    set { princEndLbl.Text = value; }
        //}
        #endregion

        public Row(int id)
        {
            _id = id;

            Size = new Size(RowWidth, RowHeight);

            Controls.Add(paymentNumberLbl);
            Controls.Add(paymentDateLbl);
            Controls.Add(interestLbl);
            Controls.Add(principalLbl);
            Controls.Add(paymentLbl);
            Controls.Add(extraPaymentTb);
            Controls.Add(cumInterestLbl);
            Controls.Add(cumPrincipalLbl);
            Controls.Add(princEndLbl);

            // new row
            int x = RowX0;
            // paymentNumberLbl
            paymentNumberLbl.AutoSize = false;
            paymentNumberLbl.TextAlign = ContentAlignment.MiddleRight;
            paymentNumberLbl.Size = new Size(RowX1 - ColBuffer, LabelHeight);
            paymentNumberLbl.Location = new Point(x, RowY);
            paymentNumberLbl.Text = "";
            paymentNumberLbl.Click += new EventHandler(click);
            // paymentDateLbl
            x += RowX1;
            paymentDateLbl.AutoSize = false;
            paymentDateLbl.TextAlign = ContentAlignment.MiddleRight;
            paymentDateLbl.Size = new Size(RowX2 - ColBuffer, LabelHeight);
            paymentDateLbl.Location = new Point(x, RowY);
            paymentDateLbl.Text = "";
            paymentDateLbl.Click += new EventHandler(click);
            // interestLbl
            x += RowX2;
            interestLbl.AutoSize = false;
            interestLbl.TextAlign = ContentAlignment.MiddleRight;
            interestLbl.Size = new Size(RowX3 - ColBuffer, LabelHeight);
            interestLbl.Location = new Point(x, RowY);
            interestLbl.Text = "";
            interestLbl.Click += new EventHandler(click);
            // principalLbl
            x += RowX3;
            principalLbl.AutoSize = false;
            principalLbl.TextAlign = ContentAlignment.MiddleRight;
            principalLbl.Size = new Size(RowX4 - ColBuffer, LabelHeight);
            principalLbl.Location = new Point(x, RowY);
            principalLbl.Text = "";
            principalLbl.Click += new EventHandler(click);
            // paymentLbl
            x += RowX4;
            paymentLbl.AutoSize = false;
            paymentLbl.TextAlign = ContentAlignment.MiddleRight;
            paymentLbl.Size = new Size(RowX5 - ColBuffer, LabelHeight);
            paymentLbl.Location = new Point(x, RowY);
            paymentLbl.Text = "";
            paymentLbl.Click += new EventHandler(click);
            // extraPaymentTb
            x += RowX5;
            extraPaymentTb.Location = new Point(x, RowY - 3);
            extraPaymentTb.Size = new Size(85, 20);
            extraPaymentTb.Text = "";
            extraPaymentTb.Submitted += new EventHandler(extraPaymentTb_Submitted);
            // cumInterestLbl
            x += RowX6;
            cumInterestLbl.AutoSize = false;
            cumInterestLbl.TextAlign = ContentAlignment.MiddleRight;
            cumInterestLbl.Size = new Size(RowX7 - ColBuffer, LabelHeight);
            cumInterestLbl.Location = new Point(x, RowY);
            cumInterestLbl.Text = "";
            cumInterestLbl.Click += new EventHandler(click);
            // cumPrincipalLbl
            x += RowX7;
            cumPrincipalLbl.AutoSize = false;
            cumPrincipalLbl.TextAlign = ContentAlignment.MiddleRight;
            cumPrincipalLbl.Size = new Size(RowX8 - ColBuffer, LabelHeight);
            cumPrincipalLbl.Location = new Point(x, RowY);
            cumPrincipalLbl.Text = "";
            cumPrincipalLbl.Click += new EventHandler(click);
            // princEndLbl
            x += RowX8;
            princEndLbl.AutoSize = false;
            princEndLbl.TextAlign = ContentAlignment.MiddleRight;
            princEndLbl.Size = new Size(RowWidth - x, LabelHeight);
            princEndLbl.Location = new Point(x, RowY);
            princEndLbl.Text = "";
            princEndLbl.Click += new EventHandler(click);
        }

        public void SetAll(SinglePayment payment)
        {
            SetAll(payment.PaymentNumber, payment.PaymentDate, payment.Interest,
                payment.Principal, payment.PaymentAmount, payment.ExtraPayment,
                payment.CumulativeInterest, payment.CumulativePrincipal,
                payment.PrincipalAtEndOfPeriod);
        }
        public void SetAll(int paymentNumber, DateTime paymentDate, double interest,
            double principal, double payment, double extraPayment, double cumInterest,
            double cumPrincipal, double princEnd)
        {
            paymentNumberLbl.Text = paymentNumber.ToString();
            paymentDateLbl.Text = String.Format("{0:MM/dd/yy}", paymentDate);
            interestLbl.Text = String.Format("{0:c}", interest);
            principalLbl.Text = String.Format("{0:c}", principal);
            paymentLbl.Text = String.Format("{0:c}", payment);
            extraPaymentTb.Text = String.Format("{0:c}", extraPayment);
            cumInterestLbl.Text = String.Format("{0:c}", cumInterest);
            cumPrincipalLbl.Text = String.Format("{0:c}", cumPrincipal);
            princEndLbl.Text = String.Format("{0:c}", princEnd);
        }

        private void click(object sender, EventArgs e)
        {
            this.Parent.Focus();
        }

        private void extraPaymentTb_Submitted(object sender, EventArgs e)
        {
            OnTextBoxChanged();
        }
        protected virtual void OnTextBoxChanged()
        {
            if (TextBoxChanged != null)
                TextBoxChanged(this, new IntEventArgs(_id));
        }
        public event IntEventHandler TextBoxChanged;
    }
}